@@ -1,7 +1,8 @@
 """Library to transport ISO7816-4 APDUs over anything"""
 import io
+import logging
 
-__version__ = '1.2.4'
+__version__ = '1.2.5'
 __title__ = 'pysatl'
 __description__ = 'Library to transport ISO7816-4 APDUs over anything'
 __long_description__ = """
@@ -300,8 +301,8 @@ class CAPDU(object):
         # CASE3S: CLA INS P1 P2 LC DATA            (LC  from   01 to   FF                       )
         # CASE3E: CLA INS P1 P2 00 LC2 DATA        (LC2 from 0001 to FFFF                       )
         # CASE4S: CLA INS P1 P2 LC DATA LE         (LC  from   01 to   FF  LE2 from 0001 to 0000)
-        # CASE4E: CLA INS P1 P2 LC DATA 00 LE2     (LC  from   01 to   FF, LE2 from 0001 to 0000)
-        # CASE4E: CLA INS P1 P2 00 LC2 DATA LE     (LC2 from 0001 to FFFF, LE  from 01   to   FF)
+        # CASE4E: CLA INS P1 P2 LC DATA 00 LE2     (LC  from   01 to   FF, LE2 from 0001 to 0000) #not standard but tolerated
+        # CASE4E: CLA INS P1 P2 00 LC2 DATA LE     (LC2 from 0001 to FFFF, LE  from 01   to   FF) #not standard but tolerated
         # CASE4E: CLA INS P1 P2 00 LC2 DATA 00 LE2 (LC2 from 0001 to FFFF, LE2 from 0001 to 0000)
         apdu_len=len(apdu_bytes)
         abytes = io.BytesIO(apdu_bytes)
@@ -366,8 +367,9 @@ class CAPDU(object):
         out.append(self.P2)
         LC = len(self.DATA)
         LE = self.LE
+        case4e = (LC>0) and (LE>0) and ((LC > 0xFF) or (LE > 0x100))
         if LC > 0:
-            if LC > 0xFF:
+            if (LC > 0xFF) or case4e :
                 out.append(0x00)
                 out += LC.to_bytes(2, byteorder='big')
             else:
@@ -375,16 +377,19 @@ class CAPDU(object):
             out += self.DATA
         if LE > 0:
             if LE == 0x10000:
+                if not case4e:
+                    out.append(0x00)
                 out.append(0x00)
                 out.append(0x00)
-                out.append(0x00)
-            elif LE > 0x100:
-                out.append(0x00)
+            elif LE > 0x100 or case4e:
+                if not case4e:
+                    out.append(0x00)
                 out += LE.to_bytes(2, byteorder='big')
-            elif LE == 0x100:
-                out.append(0x00)
             else:
-                out += LE.to_bytes(1, byteorder='big')
+                if LE == 0x100:
+                    out += bytearray([0])
+                else:
+                    out += LE.to_bytes(1, byteorder='big')
         return out
 
     def to_bytes(self):
@@ -397,7 +402,7 @@ class CAPDU(object):
         b = Utils.ba(hexstr)
         return CAPDU.from_bytes(b)
 
-    def to_hexstr(self,*,skip_long_data=False):
+    def to_hexstr(self,*,skip_long_data=False, separator=""):
         """Convert to a string of hex digits"""
         b = self.to_ba()
         header_len = 4
@@ -406,11 +411,11 @@ class CAPDU(object):
             header_len = 5
             if lc>255:
                 header_len = 7
-        dat = Utils.hexstr(b[:header_len], separator="")
+        dat = Utils.hexstr(b[:header_len], separator=separator)
         if lc:
-            dat += Utils.hexstr(b[header_len:header_len+lc], separator="",skip_long_data=skip_long_data)
+            dat += separator + Utils.hexstr(b[header_len:header_len+lc], separator=separator,skip_long_data=skip_long_data)
         if self.LE:
-            dat += Utils.hexstr(b[header_len+lc:], separator="")
+            dat += separator + Utils.hexstr(b[header_len+lc:], separator=separator)
         return dat
 
 class RAPDU(object):
@@ -463,6 +468,44 @@ class RAPDU(object):
         """Convert to bytes"""
         return bytes(self.to_ba())
 
+    def swBytes(self):
+        """SW1 and SW2 as bytearray"""
+        out = bytearray()
+        out.append(self.SW1)
+        out.append(self.SW2)
+        return out
+
+    def matchSW(self, swPat):
+        """Check if Status Word match a given hexadecimal pattern, 'X' match anything"""
+        swInt = (self.SW1 << 8) | self.SW2
+        swStr = format(swInt, 'x')
+        swPat = swPat.lower()
+        for i in range(0,len(swPat)):
+            expected = swPat[i]
+            if expected == 'x':
+                continue
+            d = swStr[i]
+            if d != expected:
+                return False
+        return True
+
+    def matchDATA(self, dataPat):
+        """Check if DATA match a given hexadecimal pattern, 'X' match anything"""
+        dataPat = dataPat.lower()
+        for i in range(0,len(dataPat)):
+            expected = dataPat[i]
+            if expected == 'x':
+                continue
+            b = self.DATA[i>>1]
+            if 0 == (i % 2):
+                d = (b >> 4) & 0xF
+            else:
+                d = b & 0xF
+            if d != int(expected,16):
+                logging.debug("Mismatch at char %d: expected %x, got %x"%(i,expected,d))
+                return False
+        return True
+
     @staticmethod
     def from_hexstr(hexstr):
         """Convert a string of hex digits to RAPDU"""
